from importlib import import_module


EMPTY = object()
def getdefaultattr(obj, name, default=None, factory=EMPTY):
    '''
    Gets the given attribute from the object,
    creating it with a default or by calling
    a factory if needed.
    '''
    try:
        return getattr(obj, name)
    except AttributeError:
        pass
    val = factory() if factory is not EMPTY else None
    setattr(obj, name, val)
    return val


def qualified_name(obj):
    '''Returns the fully-qualified name of the given object'''
    if not hasattr(obj, '__module__'):
        obj = obj.__class__
    module = obj.__module__
    if module is None or module == str.__class__.__module__:
        return obj.__qualname__
    return '{}.{}'.format(module, obj.__qualname__)


def import_qualified(name):
    '''
    Imports a fully-qualified name from a module:

        cls = import_qualified('homepage.views.index.MyForm')

    Raises an ImportError if it can't be ipmorted.
    '''
    parts = name.rsplit('.', 1)
    if len(parts) != 2:
        raise ImportError('Invalid fully-qualified name: {}'.format(name))
    try:
        return getattr(import_module(parts[0]), parts[1])
    except AttributeError:
        raise ImportError('{} not found in module {}'.format(parts[1], parts[0]))



from django.apps import apps
from django.utils.html import mark_safe
from mako.exceptions import RichTraceback
from mako.template import Template
import mako.runtime

import io
import os, os.path


def template_inheritance(obj):
    '''
    Generator that iterates the template and its ancestors.
    The order is from most specialized (furthest descendant) to
    most general (furthest ancestor).

    obj can be either:
        1. Mako Template object
        2. Mako `self` object (available within a rendering template)
    '''
    if isinstance(obj, Template):
        obj = create_mako_context(obj)['self']
    elif isinstance(obj, mako.runtime.Context):
        obj = obj['self']
    while obj is not None:
        yield obj.template
        obj = obj.inherits


def create_mako_context(template_obj, **kwargs):
    # I'm hacking into private Mako methods here, but I can't see another
    # way to do this.  Hopefully this can be rectified at some point.
    kwargs.pop('self', None)  # some contexts have self in them, and it messes up render_unicode below because we get two selfs
    runtime_context = mako.runtime.Context(io.StringIO(), **kwargs)
    runtime_context._set_with_template(template_obj)
    _, mako_context = mako.runtime._populate_self_namespace(runtime_context, template_obj)
    return mako_context


def get_template_debug_info(error):
    '''
    Retrieves mako template information for Django's debug stack trace template.
    This allows Django to print Mako syntax/runtime template errors the same way
    it prints all other errors.
    '''
    from ..convenience import get_template_loader_for_path
    dmp = apps.get_app_config('django_mako_plus')
    loader = get_template_loader_for_path(os.path.join(dmp.path, 'templates'))
    # going straight to the mako template in case the error bubbles into our
    # adapter (which calls this function and would cause infinite loops)
    template = loader.get_mako_template('stack_trace.html')
    tback = RichTraceback(error, error.__traceback__)
    return {
        'message': '',
        'source_lines': [
            ( '', mark_safe(template.render_unicode(tback=tback)) ),
        ],
        'before': '',
        'during': '',
        'after': '',
        'top': 0,
        'bottom': 0,
        'total': 0,
        'line': tback.lineno or 0,
        'name': '',
        'start': 0,
        'end': 0,
    }



from django.apps import apps
from django.conf import settings
from django.template import TemplateDoesNotExist, TemplateSyntaxError

from mako.exceptions import CompileException, SyntaxException, TemplateLookupException, TopLevelLookupException
from mako.lookup import TemplateLookup
from mako.template import Template

from .lexer import DMPLexer
from .adapter import MakoTemplateAdapter

import os
import os.path


class DMPTemplateLookup(TemplateLookup):
    '''Small extension to Mako's template lookup to provide a link back to the MakoTemplateLoader'''
    def __init__(self, template_loader, *args, **kwargs):
        super(DMPTemplateLookup, self).__init__(*args, **kwargs)
        self.template_loader = template_loader


class MakoTemplateLoader(object):
    '''Finds Mako templates for a Django app.'''
    def __init__(self, app_path, template_subdir='templates'):
        '''
        The loader looks in the app_path/templates directory unless
        the template_subdir parameter overrides this default.

        You should not normally create this object because it bypasses
        the DMP cache.  Instead, call get_template_loader() or
        get_template_loader_for_path().
        '''
        self.app_path = app_path
        dmp = apps.get_app_config('django_mako_plus')

        # calculate the template directory and check that it exists
        if template_subdir is None:  # None skips adding the template_subdir
            self.template_dir = os.path.abspath(app_path)
        else:
            self.template_dir = os.path.abspath(os.path.join(app_path, template_subdir))

        # I used to check for the existence of the template dir here, but it caused error
        # checking at engine load time (too soon).  I now wait until get_template() is called,
        # which fails with a TemplateDoesNotExist exception if the template_dir doesn't exist.

        # calculate the cache root and template search directories
        self.cache_root = os.path.join(self.template_dir, dmp.options['TEMPLATES_CACHE_DIR'])
        self.template_search_dirs = [ self.template_dir ]
        self.template_search_dirs.extend(dmp.options['TEMPLATES_DIRS'])
        # Mako doesn't allow parent directory inheritance, such as <%inherit file="../../otherapp/templates/base.html"/>
        # including the project base directory allows this through "absolute" like <%inherit file="/otherapp/templates/base.html"/>
        # (note the leading slash, which means BASE_DIR)
        self.template_search_dirs.append(settings.BASE_DIR)

        # create the actual Mako TemplateLookup, which does the actual work
        self.tlookup = DMPTemplateLookup(
            template_loader=self,
            directories=self.template_search_dirs,
            imports=dmp.template_imports,
            module_directory=self.cache_root,
            collection_size=2000,
            filesystem_checks=settings.DEBUG,
            input_encoding=dmp.options['DEFAULT_TEMPLATE_ENCODING'],
            default_filters=[],  # shouldn't be None because that causes Mako to add an html filter and override DMP's html_filter
            lexer_cls=DMPLexer,
        )


    def get_template(self, template, def_name=None):
        '''Retrieve a *Django* API template object for the given template name, using the app_path and template_subdir
           settings in this object.  This method still uses the corresponding Mako template and engine, but it
           gives a Django API wrapper around it so you can use it the same as any Django template.

           If def_name is provided, template rendering will be limited to the named def/block (see Mako docs).

           This method corresponds to the Django templating system API.
           A Django exception is raised if the template is not found or cannot compile.
        '''
        try:
            # wrap the mako template in an adapter that gives the Django template API
            return MakoTemplateAdapter(self.get_mako_template(template), def_name)
        except (TopLevelLookupException, TemplateLookupException) as e: # Mako exception raised
            raise TemplateDoesNotExist('Template "%s" not found in search path: %s.' % (template, self.template_search_dirs))
        except (CompileException, SyntaxException) as e: # Mako exception raised
            raise TemplateSyntaxError('Template "%s" raised an error: %s' % (template, e))


    def get_mako_template(self, template, force=False):
        '''Retrieve the real *Mako* template object for the given template name without any wrapper,
           using the app_path and template_subdir settings in this object.

           This method is an alternative to get_template().  Use it when you need the actual Mako template object.
           This method raises a Mako exception if the template is not found or cannot compile.

           If force is True, an empty Mako template will be created when the file does not exist.
           This option is used by the providers part of DMP and normally be left False.
        '''
        if template is None:
            raise TemplateLookupException('Template "%s" not found in search path: %s.' % (template, self.template_search_dirs))
        # get the template
        try:
            template_obj = self.tlookup.get_template(template)
        except TemplateLookupException:
            if not force:
                raise
            template_obj = Template('', filename=os.path.join(self.template_dir, template))

        # get the template
        return template_obj



from django.apps import apps
from django.core.management.base import BaseCommand, CommandError
from django.conf import settings
from django_mako_plus.management.mixins import DMPCommandMixIn

import os, os.path, shutil




class Command(DMPCommandMixIn, BaseCommand):
    help = 'Removes compiled template cache folders in your DMP-enabled app directories.'
    can_import_settings = True


    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            'appname',
            type=str,
            nargs='*',
            help='The name of one or more DMP apps. If omitted, all DMP apps are processed.'
        )
        parser.add_argument(
            '--trial-run',
            action='store_true',
            dest='trial_run',
            default=False,
            help='Display the folders that would be removed without actually removing them.'
        )


    def handle(self, *args, **options):
        dmp = apps.get_app_config('django_mako_plus')

        # save the options for later
        if options.get('trial_run'):
            self.message("Trial run: dmp_cleanup would have deleted the following folders:", level=1)

        # ensure we have a base directory
        try:
            if not os.path.isdir(os.path.abspath(settings.BASE_DIR)):
                raise CommandError('Your settings.py BASE_DIR setting is not a valid directory.  Please check your settings.py file for the BASE_DIR variable.')
        except AttributeError as e:
            print(e)
            raise CommandError('Your settings.py file is missing the BASE_DIR setting.')

        # check each dmp-enabled app
        for config in dmp.get_registered_apps():
            self.message('Cleaning up app: {}'.format(config.name), level=1)
            for subdir in ( 'templates', 'scripts', 'styles' ):
                self.deep_clean(os.path.join(config.path, subdir), options.get('trial_run'))

    def deep_clean(self, path, trial_run):
        dmp = apps.get_app_config('django_mako_plus')
        if not os.path.exists(path):
            return

        # clean this directory
        cache_dir = os.path.join(path, dmp.options['TEMPLATES_CACHE_DIR'])
        if os.path.exists(cache_dir):
            self.message('Removing {}'.format(pretty_relpath(cache_dir, settings.BASE_DIR)), level=2)
            if not trial_run:
                shutil.rmtree(cache_dir)
        else:
            self.message('Skipping {} because it does not exist'.format(pretty_relpath(cache_dir, settings.BASE_DIR)), level=2)

        # recurse to subdirectories
        for name in os.listdir(path):
            subpath = os.path.join(path, name)
            if os.path.isdir(subpath):
                self.deep_clean(subpath, trial_run)





#####################################################
###   Utility functions



def pretty_relpath(path, start):
    '''
    Returns a relative path, but only if it doesn't start with a non-pretty parent directory ".."
    '''
    relpath = os.path.relpath(path, start)
    if relpath.startswith('..'):
        return path
    return relpath



from django.apps import apps
from django.conf import settings
try:
    from django.urls import re_path              # Django 2.x
except ImportError:
    from django.conf.urls import url as re_path  # Django 1.x
from django.views.static import serve
from .router import app_resolver
import os, os.path


#########################################################
###   The default DMP url patterns
###
###   FYI, even though the valid python identifier is [_A-Za-z][_a-zA-Z0-9]*,
###   I'm simplifying it to [_a-zA-Z0-9]+ because it works for our purposes

app_name = 'django_mako_plus'
dmp = apps.get_app_config('django_mako_plus')
urlpatterns = []

# start with the DMP web files - for development time
# at production, serve this directly with Nginx/IIS/etc. instead
# there is no "if debug mode" statement here because the web server will serve the file at production before urls.py happens,
# but if this deployment step isn't done right, it will still work through this link
urlpatterns.append(re_path(
    r'^django_mako_plus/(?P<path>[^/]+)',
    serve,
    { 'document_root': os.path.join(apps.get_app_config('django_mako_plus').path, 'webroot') },
    name='DMP webroot (for devel)',
))

# add a DMP-style resolver for each app in the project directory
for config in apps.get_app_configs():
    if os.path.samefile(os.path.dirname(config.path), settings.BASE_DIR):
        urlpatterns.append(app_resolver(config.name))

# add a DMP-style resolver for the default app
if dmp.options['DEFAULT_APP']:
    try:
        apps.get_app_config(dmp.options['DEFAULT_APP'])
        urlpatterns.append(app_resolver())
    except LookupError:
        pass  # the default app in dmp's TEMPLATES entry isn't an installed app, so skip it



from django.apps import apps, AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.template import engines

from .defaults import DEFAULT_OPTIONS
from .provider.runner import ProviderRun
from .signals import dmp_signal_register_app

import threading


class Config(AppConfig):
    name = 'django_mako_plus'
    label = 'django_mako_plus'
    verbose_name = 'Django Mako Plus Templating Engine'

    def ready(self):
        '''Called by Django when the app is ready for use.'''
        # set up the options
        self.options = {}
        self.options.update(DEFAULT_OPTIONS)
        for template_engine in settings.TEMPLATES:
            if template_engine.get('BACKEND', '').startswith('django_mako_plus'):
                self.options.update(template_engine.get('OPTIONS', {}))

        # dmp-enabled apps registry
        self.registration_lock = threading.RLock()
        self.registered_apps = {}

        # init the template engine
        self.engine = engines['django_mako_plus']

        # default imports on every compiled template
        self.template_imports = [
            'import django_mako_plus',
            'import django.utils.html',     # used in template.py
        ]
        self.template_imports.extend(self.options['DEFAULT_TEMPLATE_IMPORTS'])

        # initialize the list of providers
        ProviderRun.initialize_providers()

        # set up the parameter converters (can't import until apps are set up)
        from .converter.base import ParameterConverter
        ParameterConverter._sort_converters(app_ready=True)


    def register_app(self, app=None):
        '''
        Registers an app as a "DMP-enabled" app.  Normally, DMP does this
        automatically when included in urls.py.

        If app is None, the DEFAULT_APP is registered.
        '''
        app = app or self.options['DEFAULT_APP']
        if not app:
            raise ImproperlyConfigured('An app name is required because DEFAULT_APP is empty - please use a '
                                        'valid app name or set the DEFAULT_APP in settings')
        if isinstance(app, str):
            app = apps.get_app_config(app)

        # since this only runs at startup, this lock doesn't affect performance
        with self.registration_lock:
            # short circuit if already registered
            if app.name in self.registered_apps:
                return

            # first time for this app, so add to our dictionary
            self.registered_apps[app.name] = app

            # set up the template, script, and style renderers
            # these create and cache just by accessing them
            self.engine.get_template_loader(app, 'templates', create=True)
            self.engine.get_template_loader(app, 'scripts', create=True)
            self.engine.get_template_loader(app, 'styles', create=True)

            # send the registration signal
            if self.options['SIGNALS']:
                dmp_signal_register_app.send(sender=self, app_config=app)


    def get_registered_apps(self):
        '''Returns a sequence of apps that are registered with DMP'''
        return self.registered_apps.values()


    def is_registered_app(self, app):
        '''Returns true if the given app/app name is registered with DMP'''
        if app is None:
            return False
        if isinstance(app, AppConfig):
            app = app.name
        return app in self.registered_apps



from .base import ParameterConverter

###  Decorator that denotes a converter function  ###

def parameter_converter(*convert_types):
    '''
    Decorator that denotes a function as a url parameter converter.
    '''
    def inner(func):
        for ct in convert_types:
            ParameterConverter._register_converter(func, ct)
        return func
    return inner



from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.forms.utils import flatatt
import logging
import os
from ..util import crc32, getdefaultattr
from ..util import log
from .base import BaseProvider


#####################################################
###   LinkProvider abstract base class

class LinkProvider(BaseProvider):
    '''
    Renders links like <link> and <script> based on the name of the template
    and supertemplates.
    '''
    DEFAULT_OPTIONS = {
        # explicitly sets the path to search for - if this filepath exists, DMP
        # includes a link to it in the template. globs are not supported because this
        # should resolve to one exact file. possible values:
        #   1. None: a default path is used, such as "{app}/{subdir}/{filename.ext}", prefixed
        #      with the static root at production; see subclasses for their default filenames.
        #   2. function, lambda, or other callable: called as func(provider) and
        #      should return a string
        #   3. str: used directly
        'filepath': None,

        # explicitly sets the link to be inserted into the template (when filepath exists).
        # possible values:
        #   1. None: a default path is used, such as "<script src="..."></script>, prefixed
        #      with the static url at production; see subclasses for the default link format.
        #   2. function, lambda, or other callable: called as func(provider) and
        #      should return a string
        #   3. str: inserted directly into the template
        'link': None,

        # if a template is rendered more than once in a request, should we link more than once?
        # defaults are: css=False, js=True, bundled_js=False
        'skip_duplicates': False,
    }

    def __init__(self, template, options):
        super().__init__(template, options)
        self.filepath = os.path.join(
            settings.BASE_DIR if settings.DEBUG else settings.STATIC_ROOT,
            self.build_source_filepath()
        )
        # file time and version hash
        try:
            self.mtime = int(os.stat(self.filepath).st_mtime)
            # version_id combines current time and the CRC32 checksum of file bytes
            self.version_id = (self.mtime << 32) | crc32(self.filepath)
        except FileNotFoundError:
            self.mtime = 0
            self.version_id = 0
        if log.isEnabledFor(logging.DEBUG):
            log.debug('%s created for %s: [%s]', repr(self), self.filepath, 'will link' if self.mtime > 0 else 'will skip nonexistent file')


    ### Source Filepath Building Methods ###

    def build_source_filepath(self):
        # if defined in settings, run the function or return the string
        if self.options['filepath'] is not None:
            return self.options['filepath'](self) if callable(self.options['filepath']) else self.options['filepath']
        # build the default
        if self.app_config is None:
            log.warn('{} skipped: template %s not in project subdir and `targetpath` not in settings', (self.__class__.__qualname__, self.template_relpath))
        return self.build_default_filepath()

    def build_default_filepath(self):
        # this method is overridden in CssLinkProvider and JsLinkProvider lower in this file
        raise ImproperlyConfigured('{} must set `filepath` in options (or a subclass can override build_default_filepath).'.format(self.__class__.__qualname__))


    ### Target Link Building Methods ###

    def build_target_link(self, provider_run, data):
        # if defined in settings, run the function or return the string
        if self.options['link'] is not None:
            return self.options['link'](self) if callable(self.options['link']) else self.options['link']
        # build the default
        if self.app_config is None:
            log.warn('{} skipped: template %s not in project subdir and `targetpath` not in settings', (self.__class__.__qualname__, self.template_relpath))
        return self.build_default_link(provider_run, data)

    def build_default_link(self, provider_run, data):
        # this method is overridden in CssLinkProvider and JsLinkProvider lower in this file
        raise ImproperlyConfigured('{} must set `link` in options (or a subclass can override build_default_link).'.format(self.__class__.__qualname__))


    ### Provider Run Methods ###

    def start(self, provider_run, data):
        # add a set to the request (fallback to provider_run if request is None) for skipping duplicates
        if self.options['skip_duplicates']:
            data['seen'] = getdefaultattr(
                provider_run.request.dmp if provider_run.request is not None else provider_run,
                '_LinkProvider_Filename_Cache_',
                factory=set,
            )
        # enabled providers in the chain go here
        data['enabled'] = []

    def provide(self, provider_run, data):
        filepath = self.filepath
        # delaying printing of tag to finish() because the JsContextProvider delays and this must go after it
        # short circuit if the file for this provider doesn't exist
        if self.mtime == 0:
            return
        # short circut if we're skipping duplicates and we've already seen this one
        if self.options['skip_duplicates']:
            if filepath in data['seen']:
                if log.isEnabledFor(logging.DEBUG):
                    log.debug('%s skipped duplicate %s', repr(self), self.filepath)
                return
            data['seen'].add(filepath)
        # if we get here, this provider is enabled, so add it to the list
        if log.isEnabledFor(logging.DEBUG):
            log.debug('%s linking %s', repr(self), self.filepath)
        data['enabled'].append(self)

    def finish(self, provider_run, data):
        for provider in data['enabled']:
            provider_run.write(provider.build_target_link(provider_run, data))


#####################################
###  CssLinkProvider

class CssLinkProvider(LinkProvider):
    '''Generates a CSS <link>'''
    DEFAULT_OPTIONS = {
        'group': 'styles',
        # if a template is rendered more than once in a request, we usually don't
        # need to include the css again.
        'skip_duplicates': True,
    }

    def build_default_filepath(self):
        '''Called when 'filepath' is not defined in the settings'''
        return os.path.join(
            self.app_config.name,
            'styles',
            self.template_relpath + '.css',
        )

    def build_default_link(self, provider_run, data):
        '''Called when 'link' is not defined in the settings'''
        attrs = {}
        attrs['rel'] = 'stylesheet'
        attrs["data-context"] = provider_run.uid
        attrs["href"] ="{}?{:x}".format(
            os.path.join(
                settings.STATIC_URL,
                os.path.relpath(
                    self.build_source_filepath(),
                    settings.BASE_DIR,
                ),
            ).replace(os.path.sep, '/'),
            self.version_id,
        )
        return '<link{} />'.format(flatatt(attrs))


##############################################
###   JsLinkProvider

class JsLinkProvider(LinkProvider):
    '''Generates a JS <script>.'''
    DEFAULT_OPTIONS = {
        'group': 'scripts',
        # if a template is rendered more than once in a request, we should link each one
        # so the script runs again each time the template runs
        'skip_duplicates': False,
        # whether to create an async script tag
        'async': False,
    }

    def build_default_filepath(self):
        '''Called when 'filepath' is not defined in the settings'''
        return os.path.join(
            self.app_config.name,
            'scripts',
            self.template_relpath + '.js',
        )

    def build_default_link(self, provider_run, data):
        '''Called when 'link' is not defined in the settings'''
        attrs = {}
        attrs["data-context"] = provider_run.uid
        attrs["src"] ="{}?{:x}".format(
            os.path.join(
                settings.STATIC_URL,
                os.path.relpath(
                    self.build_source_filepath(),
                    settings.BASE_DIR,
                ),
            ).replace(os.path.sep, '/'),
            self.version_id,
        )
        if self.options['async']:
            attrs['async'] = "true"
        return '<script{}></script>'.format(flatatt(attrs))




################################################################
###   Special type of list used for url params

class URLParamList(list):
    '''
    A simple extension to Python's list that returns '' for indices that don't exist.
    For example, if the object is ['a', 'b'] and you call obj[5], it will return ''
    rather than throwing an IndexError.  This makes dealing with url parameters
    simpler since you don't have to check the length of the list.
    '''
    def __getitem__(self, idx):
        '''Returns the element at idx, or '' if idx is beyond the length of the list'''
        return self.get(idx, '')

    def get(self, idx, default=''):
        '''Returns the element at idx, or default if idx is beyond the length of the list'''
        # if the index is beyond the length of the list, return ''
        if isinstance(idx, int) and (idx >= len(self) or idx < -1 * len(self)):
            return default
        # else do the regular list function (for int, slice types, etc.)
        return super().__getitem__(idx)



from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ViewDoesNotExist
from django.template import TemplateDoesNotExist
from django.views.generic import View

from .decorators import view_function, CONVERTER_ATTRIBUTE_NAME
from ..util import import_qualified, log

import inspect
import threading
from importlib import import_module
from importlib.util import find_spec



########################################################
###   Cached routers

CACHED_VIEW_FUNCTIONS = {}
rlock = threading.RLock()


def get_view_function(module_name, function_name, fallback_app=None, fallback_template=None, verify_decorator=True):
    '''
    Retrieves a view function from the cache, finding it if the first time.
    Raises ViewDoesNotExist if not found.  This is called by resolver.py.
    '''
    # first check the cache (without doing locks)
    key = ( module_name, function_name )
    try:
        return CACHED_VIEW_FUNCTIONS[key]
    except KeyError:
        with rlock:
            # try again now that we're locked
            try:
                return CACHED_VIEW_FUNCTIONS[key]
            except KeyError:
                # if we get here, we need to load the view function
                func = find_view_function(module_name, function_name, fallback_app, fallback_template, verify_decorator)
                # cache in production mode
                if not settings.DEBUG:
                    CACHED_VIEW_FUNCTIONS[key] = func
                return func

    # the code should never be able to get here
    raise Exception("Django-Mako-Plus error: get_view_function() should not have been able to get to this point.  Please notify the owner of the DMP project.  Thanks.")


def find_view_function(module_name, function_name, fallback_app=None, fallback_template=None, verify_decorator=True):
    '''
    Finds a view function, class-based view, or template view.
    Raises ViewDoesNotExist if not found.
    '''
    dmp = apps.get_app_config('django_mako_plus')

    # I'm first calling find_spec first here beacuse I don't want import_module in
    # a try/except -- there are lots of reasons that importing can fail, and I just want to
    # know whether the file actually exists.  find_spec raises AttributeError if not found.
    try:
        spec = find_spec(module_name)
    except ValueError:
        spec = None
    if spec is None:
        # no view module, so create a view function that directly renders the template
        try:
            return create_view_for_template(fallback_app, fallback_template)
        except TemplateDoesNotExist as e:
            raise ViewDoesNotExist('view module {} not found, and fallback template {} could not be loaded ({})'.format(module_name, fallback_template, e))

    # load the module and function
    try:
        module = import_module(module_name)
        func = getattr(module, function_name)
        func.view_type = 'function'
    except ImportError as e:
        raise ViewDoesNotExist('module "{}" could not be imported: {}'.format(module_name, e))
    except AttributeError as e:
        raise ViewDoesNotExist('module "{}" found successfully, but "{}" was not found: {}'.format(module_name, function_name, e))

    # if class-based view, call as_view() to get a view function to it
    if inspect.isclass(func) and issubclass(func, View):
        func = func.as_view()
        func.view_type = 'class'

    # if regular view function, check the decorator
    elif verify_decorator and not view_function.is_decorated(func):
        raise ViewDoesNotExist("view {}.{} was found successfully, but it must be decorated with @view_function or be a subclass of django.views.generic.View.".format(module_name, function_name))

    # attach a converter to the view function
    if dmp.options['PARAMETER_CONVERTER'] is not None:
        try:
            converter = import_qualified(dmp.options['PARAMETER_CONVERTER'])(func)
            setattr(func, CONVERTER_ATTRIBUTE_NAME, converter)
        except ImportError as e:
            raise ImproperlyConfigured('Cannot find PARAMETER_CONVERTER: {}'.format(str(e)))

    # return the function/class
    return func


def create_view_for_template(app_name, template_name):
    '''
    Creates a view function for templates (used whe a view.py file doesn't exist but the .html does)
    Raises TemplateDoesNotExist if the template doesn't exist.
    '''
    # ensure the template exists
    apps.get_app_config('django_mako_plus').engine.get_template_loader(app_name).get_template(template_name)
    # create the view function
    def template_view(request, *args, **kwargs):
        # not caching the template object (getting it each time) because Mako has its own cache
        dmp = apps.get_app_config('django_mako_plus')
        template = dmp.engine.get_template_loader(app_name).get_template(template_name)
        return template.render_to_response(request=request, context=kwargs)
    template_view.view_type = 'template'
    return template_view



from django.apps import apps
from django.conf import settings
from django.http import HttpResponse
from django.utils.html import mark_safe
from django.template import Context, RequestContext

from mako.exceptions import html_error_template, MakoException
from mako.exceptions import RichTraceback

from ..exceptions import RedirectException
from ..signals import dmp_signal_pre_render_template, dmp_signal_post_render_template, dmp_signal_redirect_exception
from ..util import log
from .util import get_template_debug_info

import logging
import mimetypes
import os
import os.path
import sys



class MakoTemplateAdapter(object):
    '''A thin wrapper for a Mako template object that provides the Django API methods.'''
    def __init__(self, mako_template, def_name=None):
        '''
        Creates an adapter that corresponds to the Django API.

        If def_name is provided, template rendering will be limited to the named def/block (see Mako docs).
        This can also be provided in the call to render().
        '''
        self.mako_template = mako_template
        self.def_name = def_name


    @property
    def engine(self):
        '''Returns the DMP engine (method required by Django specs)'''
        dmp = apps.get_app_config('django_mako_plus')
        return dmp.engine


    def render(self, context=None, request=None, def_name=None):
        '''
        Renders a template using the Mako system.  This method signature conforms to
        the Django template API, which specifies that template.render() returns a string.

            @context  A dictionary of name=value variables to send to the template page.  This can be a real dictionary
                      or a Django Context object.
            @request  The request context from Django.  If this is None, any TEMPLATE_CONTEXT_PROCESSORS defined in your settings
                      file will be ignored but the template will otherwise render fine.
            @def_name Limits output to a specific top-level Mako <%block> or <%def> section within the template.
                      If the section is a <%def>, any parameters must be in the context dictionary.  For example,
                      def_name="foo" will call <%block name="foo"></%block> or <%def name="foo()"></def> within
                      the template.

        Returns the rendered template as a unicode string.

        The method triggers two signals:
            1. dmp_signal_pre_render_template: you can (optionally) return a new Mako Template object from a receiver to replace
               the normal template object that is used for the render operation.
            2. dmp_signal_post_render_template: you can (optionally) return a string to replace the string from the normal
               template object render.
        '''
        dmp = apps.get_app_config('django_mako_plus')
        # set up the context dictionary, which is the variables available throughout the template
        context_dict = {}
        # if request is None, add some default items because the context processors won't happen
        if request is None:
            context_dict['settings'] = settings
            context_dict['STATIC_URL'] = settings.STATIC_URL
        # let the context_processors add variables to the context.
        if not isinstance(context, Context):
            context = Context(context) if request is None else RequestContext(request, context)
        with context.bind_template(self):
            for d in context:
                context_dict.update(d)
        context_dict.pop('self', None)  # some contexts have self in them, and it messes up render_unicode below because we get two selfs

        # send the pre-render signal
        if dmp.options['SIGNALS'] and request is not None:
            for receiver, ret_template_obj in dmp_signal_pre_render_template.send(sender=self, request=request, context=context, template=self.mako_template):
                if ret_template_obj is not None:
                    if isinstance(ret_template_obj, MakoTemplateAdapter):
                        self.mako_template = ret_template_obj.mako_template   # if the signal function sends a MakoTemplateAdapter back, use the real mako template inside of it
                    else:
                        self.mako_template = ret_template_obj                 # if something else, we assume it is a mako.template.Template, so use it as the template

        # do we need to limit down to a specific def?
        # this only finds within the exact template (won't go up the inheritance tree)
        # I wish I could make it do so, but can't figure this out
        render_obj = self.mako_template
        if def_name is None:
            def_name = self.def_name
        if def_name:  # do we need to limit to just a def?
            render_obj = self.mako_template.get_def(def_name)

        # PRIMARY FUNCTION: render the template
        template_name = '%s::%s' % (self.mako_template.filename or 'string', def_name or 'body')
        if log.isEnabledFor(logging.INFO):
            log.info('rendering template %s', template_name)
        if settings.DEBUG:
            try:
                content = render_obj.render_unicode(**context_dict)
            except Exception as e:
                e.template_debug = get_template_debug_info(e)
                log.exception('exception raised during template rendering: %s', e)  # to the console
                raise   # reraises the same error
        else:  # this is outside the above "try" loop because in non-DEBUG mode, we want to let the exception throw out of here (without having to re-raise it)
            content = render_obj.render_unicode(**context_dict)

        # send the post-render signal
        if dmp.options['SIGNALS'] and request is not None:
            for receiver, ret_content in dmp_signal_post_render_template.send(sender=self, request=request, context=context, template=self.mako_template, content=content):
                if ret_content is not None:
                    content = ret_content  # sets it to the last non-None return in the signal receiver chain

        # return
        return mark_safe(content)


    def render_to_response(self, context=None, request=None, def_name=None, content_type=None, status=None, charset=None):
        '''
        Renders the template and returns an HttpRequest object containing its content.

        This method returns a django.http.Http404 exception if the template is not found.
        If the template raises a django_mako_plus.RedirectException, the browser is redirected to
        the given page, and a new request from the browser restarts the entire DMP routing process.
        If the template raises a django_mako_plus.InternalRedirectException, the entire DMP
        routing process is restarted internally (the browser doesn't see the redirect).

            @request      The request context from Django.  If this is None, any TEMPLATE_CONTEXT_PROCESSORS defined in your settings
                          file will be ignored but the template will otherwise render fine.
            @template     The template file path to render.  This is relative to the app_path/controller_TEMPLATES_DIR/ directory.
                          For example, to render app_path/templates/page1, set template="page1.html", assuming you have
                          set up the variables as described in the documentation above.
            @context      A dictionary of name=value variables to send to the template page.  This can be a real dictionary
                          or a Django Context object.
            @def_name     Limits output to a specific top-level Mako <%block> or <%def> section within the template.
                          For example, def_name="foo" will call <%block name="foo"></%block> or <%def name="foo()"></def> within the template.
            @content_type The MIME type of the response.  Defaults to settings.DEFAULT_CONTENT_TYPE (usually 'text/html').
            @status       The HTTP response status code.  Defaults to 200 (OK).
            @charset      The charset to encode the processed template string (the output) with.  Defaults to settings.DEFAULT_CHARSET (usually 'utf-8').

        The method triggers two signals:
            1. dmp_signal_pre_render_template: you can (optionally) return a new Mako Template object from a receiver to replace
               the normal template object that is used for the render operation.
            2. dmp_signal_post_render_template: you can (optionally) return a string to replace the string from the normal
               template object render.
        '''
        try:
            if content_type is None:
                content_type = mimetypes.types_map.get(os.path.splitext(self.mako_template.filename)[1].lower(), settings.DEFAULT_CONTENT_TYPE)
            if charset is None:
                charset = settings.DEFAULT_CHARSET
            if status is None:
                status = 200
            content = self.render(context=context, request=request, def_name=def_name)
            return HttpResponse(content.encode(charset), content_type='%s; charset=%s' % (content_type, charset), status=status)

        except RedirectException: # redirect to another page
            e = sys.exc_info()[1]
            if request is None:
                log.info('a template redirected processing to %s', e.redirect_to)
            else:
                log.info('view function %s.%s redirected processing to %s', request.dmp.module, request.dmp.function, e.redirect_to)
            # send the signal
            dmp = apps.get_app_config('django_mako_plus')
            if dmp.options['SIGNALS']:
                dmp_signal_redirect_exception.send(sender=sys.modules[__name__], request=request, exc=e)
            # send the browser the redirect command
            return e.get_response(request)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
import os
import os.path
import shutil
import collections
import logging
from .base import BaseProvider
from ..util import log
from ..command import run_command


class CompileProvider(BaseProvider):
    '''
    Runs a command, such as compiling *.scss or *.less, when an output file
    timestamp is older than the source file. In production mode, this check
    is done only once (the first time a template is run) per server start.

    When settings.DEBUG=True, checks for a recompile every request.
    When settings.DEBUG=False, checks for a recompile only once per server run.
    '''
    def __init__(self, template, options):
        super().__init__(template, options)
        self.sourcepath = os.path.join(settings.BASE_DIR if settings.DEBUG else settings.STATIC_ROOT, self.build_sourcepath())
        self.targetpath = os.path.join(settings.BASE_DIR if settings.DEBUG else settings.STATIC_ROOT, self.build_targetpath())
        # since this is in the constructor, it runs only one time per server
        # run when in production mode
        if not os.path.exists(self.sourcepath):
            msg = 'skipping nonexistent file'
        elif self.needs_compile:
            msg = 'compiling file'
            if not os.path.exists(os.path.dirname(self.targetpath)):
                os.makedirs(os.path.dirname(self.targetpath))
            run_command(*self.build_command())
        else:
            msg = 'already up to date'
        if log.isEnabledFor(logging.DEBUG):
            log.debug('%s created for %s: [%s]', repr(self), self.sourcepath, msg)

    DEFAULT_OPTIONS = {
        'group': 'styles',

        # explicitly sets the path to search for - if this filepath exists, DMP
        # includes a link to it in the template. globs are not supported because this
        # should resolve to one exact file. possible values:
        #   1. None: a default path is used, such as "{app}/{subdir}/{filename.ext}", prefixed
        #      with the static root at production; see subclasses for their default filenames.
        #   2. function, lambda, or other callable: called as func(provider) and
        #      should return a string
        #   3. str: used directly
        'sourcepath': None,

        # explicitly sets the path to search for - if this filepath exists, DMP
        # includes a link to it in the template. globs are not supported because this
        # should resolve to one exact file. possible values:
        #   1. None: a default path is used, such as "{app}/{subdir}/{filename.ext}", prefixed
        #      with the static root at production; see subclasses for their default filenames.
        #   2. function, lambda, or other callable: called as func(provider) and
        #      should return a string
        #   3. str: used directly
        'targetpath': None,

        # explicitly sets the command to be run. possible values:
        #   1. None or []: the default command is run
        #   2. function, lambda, or other callable: called as func(provider), expects list as return
        #   3. list: used directly in the call to subprocess module
        'command': [],
    }

    def build_sourcepath(self):
        # if defined in settings, run the function or return the string
        if self.options['sourcepath'] is not None:
            return self.options['sourcepath'](self) if callable(self.options['sourcepath']) else self.options['sourcepath']
        # build the default
        if self.app_config is None:
            log.warn('{} skipped: template %s not in project subdir and `targetpath` not in settings', (self.__class__.__qualname__, self.template_relpath))
        return self.build_default_sourcepath()

    def build_default_sourcepath(self):
        # this method is overridden in CompileScssProvider and CompileLessProvider lower in this file
        raise ImproperlyConfigured('{} must set `sourcepath` in options (or a subclass can override build_default_sourcepath).'.format(self.__class__.__qualname__))

    def build_targetpath(self):
        # if defined in settings, run the function or return the string
        if self.options['targetpath'] is not None:
            return self.options['targetpath'](self) if callable(self.options['targetpath']) else self.options['targetpath']
        # build the default
        if self.app_config is None:
            log.warn('{} skipped: template %s not in project subdir and `targetpath` not in settings', (self.__class__.__qualname__, self.template_relpath))
        return self.build_default_targetpath()

    def build_default_targetpath(self):
        # this method is overridden in CompileScssProvider and CompileLessProvider lower in this file
        raise ImproperlyConfigured('{} must set `targetpath` in options (or a subclass can override build_default_targetpath).'.format(self.__class__.__qualname__))

    def build_command(self):
        '''Returns the command to run, as a list (see subprocess module)'''
        # if defined in settings, run the function or return the string
        if self.options['command']:
            return self.options['command'](self) if callable(self.options['command']) else self.options['command']
        # build the default
        return self.build_default_command()

    def build_default_command(self):
        # this method is overridden in CompileScssProvider and CompileLessProvider lower in this file
        raise ImproperlyConfigured('{} must set `command` in options (or a subclass can override build_default_command).'.format(self.__class__.__qualname__))

    @property
    def needs_compile(self):
        '''Returns True if self.sourcepath is newer than self.targetpath'''
        try:
            source_mtime = os.stat(self.sourcepath).st_mtime
        except OSError:  # no source for this template, so just return
            return False
        try:
            target_mtime = os.stat(self.targetpath).st_mtime
        except OSError: # target doesn't exist, so compile
            return True
        # both source and target exist, so compile if source newer
        return source_mtime > target_mtime


###################
###   Sass

class CompileScssProvider(CompileProvider):
    '''Specialized CompileProvider for SCSS'''
    def build_default_sourcepath(self):
        return os.path.join(
            self.app_config.name,
            'styles',
            self.template_relpath + '.scss',
        )

    def build_default_targetpath(self):
        # posixpath because URLs use forward slash
        return os.path.join(
            self.app_config.name,
            'styles',
            self.template_relpath + '.css',
        )

    def build_default_command(self):
        return [
            shutil.which('sass'),
            '--source-map',
            '--load-path={}'.format(settings.BASE_DIR),
            self.sourcepath,
            self.targetpath,
        ]


#####################
###   Less

class CompileLessProvider(CompileProvider):
    '''Specialized CompileProvider that contains settings for *.less files.'''
    def build_default_sourcepath(self):
        return os.path.join(
            self.app_config.name,
            'styles',
            self.template_relpath + '.less',
        )

    def build_default_targetpath(self):
        # posixpath because URLs use forward slash
        return os.path.join(
            self.app_config.name,
            'styles',
            self.template_relpath + '.css',
        )

    def build_default_command(self):
        return [
            shutil.which('lessc'),
            '--source-map',
            self.sourcepath,
            self.targetpath,
        ]



from django.apps import apps
import os, os.path


##############################################################
###   Convenience functions
###   These are imported into __init__.py

def get_template_loader(app, subdir='templates'):
    '''
    Convenience method that calls get_template_loader() on the DMP
    template engine instance.
    '''
    dmp = apps.get_app_config('django_mako_plus')
    return dmp.engine.get_template_loader(app, subdir, create=True)


def get_template(app, template_name, subdir="templates"):
    '''
    Convenience method that retrieves a template given the app and
    name of the template.
    '''
    dmp = apps.get_app_config('django_mako_plus')
    return dmp.engine.get_template_loader(app, subdir, create=True).get_template(template_name)


def render_template(request, app, template_name, context=None, subdir="templates", def_name=None):
    '''
    Convenience method that directly renders a template, given the app and template names.
    '''
    return get_template(app, template_name, subdir).render(context, request, def_name)


def get_template_loader_for_path(path, use_cache=True):
    '''
    Convenience method that calls get_template_loader_for_path() on the DMP
    template engine instance.
    '''
    dmp = apps.get_app_config('django_mako_plus')
    return dmp.engine.get_template_loader_for_path(path, use_cache)


def get_template_for_path(path, use_cache=True):
    '''
    Convenience method that retrieves a template given a direct path to it.
    '''
    dmp = apps.get_app_config('django_mako_plus')
    app_path, template_name = os.path.split(path)
    return dmp.engine.get_template_loader_for_path(app_path, use_cache=use_cache).get_template(template_name)


def render_template_for_path(request, path, context=None, use_cache=True, def_name=None):
    '''
    Convenience method that directly renders a template, given a direct path to it.
    '''
    return get_template_for_path(path, use_cache).render(context, request, def_name)



# set up the logger
import logging
log = logging.getLogger('django_mako_plus')


# public functions
from .base58 import b58enc, b58dec
from .datastruct import merge_dicts, flatten, crc32
from .reflect import getdefaultattr, qualified_name, import_qualified



from django.utils.module_loading import import_string
import json
import logging
from ..version import __version__
from ..util import log
from .base import BaseProvider

###################################
###  JS Context Provider

class JsContextProvider(BaseProvider):
    '''
    Adds all js_context() variables to DMP_CONTEXT.
    This should be listed before JsLinkProvider so the
    context variables are available during <script> runs.
    '''
    DEFAULT_OPTIONS = {
        # the group this provider is part of.  this only matters when
        # the html page limits the providers that will be called with
        # ${ django_mako_plus.links(group="...") }
        'group': 'scripts',
        # the encoder to use for the JSON structure
        'encoder': 'django.core.serializers.json.DjangoJSONEncoder',
    }

    def __init__(self, template, options):
        super().__init__(template, options)
        self.encoder = import_string(self.options['encoder'])
        self.template = "{}/{}".format(self.app_config.name, self.template_relpath)
        if log.isEnabledFor(logging.DEBUG):
            log.debug('%s created', repr(self))

    def start(self, provider_run, data):
        data['templates'] = []

    def provide(self, provider_run, data):
        data['templates'].append(self.template)

    def finish(self, provider_run, data):
        if len(data['templates']) == 0:
            return
        context_data = {
            jscontext('__router__'): {
                'template': self.template,
                'app': provider_run.request.dmp.app if provider_run.request is not None else None,
                'page': provider_run.request.dmp.page if provider_run.request is not None else None,
                'log': log.isEnabledFor(logging.DEBUG),
            },
        }
        for k in provider_run.context.keys():
            if isinstance(k, jscontext):
                value = provider_run.context[k]
                context_data[k] = value.__jscontext__() if hasattr(value, '__jscontext__') else value
        # add to the JS DMP_CONTEXT
        provider_run.write('<script>')
        provider_run.write('DMP_CONTEXT.set("{version}", "{contextid}", {data}, {templates});'.format(
            version=__version__,
            contextid=provider_run.uid,
            data=json.dumps(context_data, cls=self.encoder, separators=(',', ':')) if context_data else '{}',
            templates=json.dumps(data['templates']),
        ))
        provider_run.write('</script>')


class jscontext(str):
    '''
    Marks a key in the context dictionary as a JS context item.
    JS context items are sent to the template like normal,
    but they are also added to the runtime JS namespace.

    See the tutorial for more information on this function.
    '''
    # no code needed, just using the class for identity



from django.conf import settings
from django.core.management import call_command
from django.forms.utils import flatatt
import os
import os.path
import posixpath
from ..util import merge_dicts
from .link import CssLinkProvider, JsLinkProvider
from ..management.commands.dmp_webpack import Command as WebpackCommand



class WebpackJsLinkProvider(JsLinkProvider):
    '''Generates a JS <script> tag for an app-level JS bundle file, if it exists.'''
    DEFAULT_OPTIONS = {
        'skip_duplicates': True,
        'async': False,
        'create_entry': True,
    }

    def build_default_filepath(self):
        # if development, recreate the entry file for this app
        if settings.DEBUG and self.options['create_entry']:
            call_command(WebpackCommand(running_inline=True), self.app_config.name, overwrite=True)

        # return the bundle path
        return os.path.join(
            self.app_config.name,
            'scripts',
            '__bundle__.js',
        )

    def build_default_link(self, provider_run, data):
        attrs = {}
        attrs["data-context"] = provider_run.uid
        attrs["src"] ="{}?{:x}".format(
            posixpath.join(
                settings.STATIC_URL,
                self.app_config.name,
                'scripts',
                '__bundle__.js',
            ),
            self.version_id,
        )
        attrs['onload'] = "DMP_CONTEXT.checkBundleLoaded('{}')".format(provider_run.uid)
        if self.options['async']:
            attrs['async'] = 'async'
        return '<script{}></script>'.format(flatatt(attrs))

    def finish(self, provider_run, data):
        # this trigger call must be separate because script links may not always
        # render (duplicates are skipped)
        super().finish(provider_run, data)
        if len(data['enabled']) > 0:
            provider_run.write('<script data-context="{uid}">DMP_CONTEXT.callBundleContext("{uid}")</script>'.format(
                uid=provider_run.uid,
            ))



import collections
import zlib


def merge_dicts(*dicts):
    '''
    Shallow merges an arbitrary number of dicts, starting
    with the first argument and updating through the
    last argument (last dict wins on conflicting keys).
    '''
    merged = {}
    for d in dicts:
        if d:
            merged.update(d)
    return merged


def flatten(*args):
    '''Generator that recursively flattens embedded lists, tuples, etc.'''
    for arg in args:
        if isinstance(arg, collections.Iterable) and not isinstance(arg, (str, bytes)):
            yield from flatten(*arg)
        else:
            yield arg



def crc32(filename):
    '''
    Calculates the CRC checksum for a file.
    Using CRC32 because security isn't the issue and don't need perfect noncollisions.
    We just need to know if a file has changed.

    On my machine, crc32 was 20 times faster than any hashlib algorithm,
    including blake and md5 algorithms.
    '''
    result = 0
    with open(filename, 'rb') as fin:
        while True:
            chunk = fin.read(48)
            if len(chunk) == 0:
                break
            result = zlib.crc32(chunk, result)
    return result




# try to import MiddlewareMixIn (Django 1.10+)
try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    # create a dummy MiddlewareMixin if older Django
    MiddlewareMixin = object

from .router import RequestViewWrapper, RoutingData




##########################################################
###   Middleware the prepares the request for
###   use with the controller.


class RequestInitMiddleware(MiddlewareMixin):
    '''
    A required middleware class that adds a RoutingData object to the request
    at the earliest possible moment.

    Note that VIEW middleware functions can not only read the RouteData variables, but they can
    adjust values as well. This power should be used with great responsibility, but it allows
    middleware to adjust the app, page, function, and url params if needed.
    '''
    # This singleton is set on the request object early in the request (during middleware).
    # Once urls.py has processed, request.dmp is changed to a populated RoutingData object.
    INITIAL_ROUTING_DATA = RoutingData()


    def process_request(self, request):
        request.dmp = self.INITIAL_ROUTING_DATA

    def process_view(self, request, view_func, view_args, view_kwargs):
        # view_func will be a RequestViewWrapper when our resolver (DMPResolver) matched
        if isinstance(view_func, RequestViewWrapper):
            view_func.routing_data.request = request
            request.dmp = view_func.routing_data



from django.apps import apps

from ..decorators import BaseDecorator
from .urlparams import URLParamList
from .discover import get_view_function

from urllib.parse import unquote


class RoutingData(object):
    '''
    The routing information for a request.  This is created during url resolution when a pattern
    matches (see resolver.py).

    During middleware, this is not available
        request.dmp.app         The Django application (such as "homepage"), as a string.
        request.dmp.page        The view module (such as "index" for index.py), as a string.
        request.dmp.function    The function within the view module to be called (usually "process_request"),
                                as a string.
        request.dmp.module      The module path in Python terms (such as homepage.views.index), as a string.
        request.dmp.callable    The view callable (function, method, etc.) to be called by the router.
        request.dmp.view_type   The type of view: function, class, or template.
        request.dmp.urlparams   A list of the remaining url parts, as a list of strings. Parameter conversion
                                uses the values in this list.

    '''
    def __init__(self, app=None, page=None, function=None, urlparams=None):
        '''These variables are set by the process_view method above'''
        # the request object is set later by the middleware so the render methods work
        self.request = None

        # period and dash cannot be in python names, but we allow dash in app, dash in page, and dash/period in function
        self.app = app.replace('-', '_') if app is not None else None
        self.page = page.replace('-', '_') if page is not None else None
        if function and function != 'process_request':
            self.function = function.replace('.', '_').replace('-', '_')
            fallback_template = '{}.{}.html'.format(page, function)
        else:
            self.function = 'process_request'
            fallback_template = '{}.html'.format(page)

        # set the module and function
        # the return of get_router_function might be a function, a class-based view, or a template
        if self.app is not None and self.page is not None:
            self.module = '.'.join([ self.app, 'views', self.page ])
            self.callable = get_view_function(self.module, self.function, self.app, fallback_template)
        else:
            self.module = None
            self.callable = None
        self.view_type = self.callable.view_type if self.callable is not None else None

        # parse the urlparams
        # note that I'm not using unquote_plus because the + switches to a space *after* the question mark (in the regular parameters)
        # in the normal url, spaces should be quoted with %20.  Thanks Rosie for the tip.
        if isinstance(urlparams, (list, tuple)):
            self.urlparams = URLParamList(urlparams)
        elif urlparams:
            self.urlparams = URLParamList(( unquote(s) for s in urlparams.split('/') ))
        else:
            self.urlparams = URLParamList()


    def __repr__(self):
        return '<RoutingData app={}, page={}, module={}, function={}, view_type={}, urlparams={}>'.format(
             self.app,
             self.page,
             self.module,
             self.function,
             self.view_type,
             self.urlparams,
        )


    def _debug(self):
        return 'django_mako_plus RoutingData:' + \
            ''.join(('\n\t{: <16}{}'.format(k, v) for k, v in (
                ( 'app', self.app ),
                ( 'page', self.page ),
                ( 'module', self.module ),
                ( 'function', self.function ),
                ( 'callable', self.callable ),
                ( 'view_type', self.view_type ),
                ( 'urlparams', self.urlparams ),
            )))


    def render(self, template, context=None, def_name=None, subdir='templates', content_type=None, status=None, charset=None):
        '''App-specific render function that renders templates in the *current app*, attached to the request for convenience'''
        if self.request is None:
            raise ValueError("RoutingData.render() can only be called after the view middleware is run. Check that `django_mako_plus.middleware` is in MIDDLEWARE.")
        dmp = apps.get_app_config('django_mako_plus')
        template_loader = dmp.engine.get_template_loader(self.app, subdir)
        template_adapter = template_loader.get_template(template)
        return getattr(template_adapter, 'render_to_response')(context=context, request=self.request, def_name=def_name, content_type=content_type, status=status, charset=charset)


    def render_to_string(self, template, context=None, def_name=None, subdir='templates'):
        '''App-specific render function that renders templates in the *current app*, attached to the request for convenience'''
        if self.request is None:
            raise ValueError("RoutingData.render() can only be called after the view middleware is run. Check that `django_mako_plus.middleware` is in MIDDLEWARE.")
        dmp = apps.get_app_config('django_mako_plus')
        template_loader = dmp.engine.get_template_loader(self.app, subdir)
        template_adapter = template_loader.get_template(template)
        return getattr(template_adapter, 'render')(context=context, request=self.request, def_name=def_name)



from django.http import HttpResponse

# this redirect key is (hopefully) unique but generic so it doesn't signpost the use of DMP/Django.
# not prefixing with X- because that's now deprecated.
REDIRECT_HEADER_KEY = 'Redirect-Location'


###############################################################################
###   Redirect with Javascript instead of 301/302
###   See also exceptions.py for two additional redirect methods

class HttpResponseJavascriptRedirect(HttpResponse):
    '''
    Sends a regular HTTP 200 OK response that contains Javascript to
    redirect the browser:

        <script>window.location.assign("...");</script>.

    If redirect_to is empty, it redirects to the current location (essentially refreshing
    the current page):

        <script>window.location.assign(window.location.href);</script>.

    Normally, redirecting should be done via HTTP 302 rather than Javascript.
    Use this class when your only choice is through Javascript.

    For example, suppose you need to redirect the top-level page from an Ajax response.
    Ajax redirects normally only redirects the Ajax itself (not the page that initiated the call),
    and this default behavior is usually what is needed.  However, there are instances when the
    entire page must be redirected, even if the call is Ajax-based.

    After the redirect_to parameter, you can use any of the normal HttpResponse constructor arguments.

    If you need to omit the surrounding <script> tags, send "include_script_tag=False" to
    the constructor. One use case for omitting the tags is when the caller is a
    JQuery $.script() ajax call.

    A custom header is set in the response.  This allows middleware, your web server, or
    calling JS code to adjust the redirect if needed.

    Note that this method doesn't use the <meta> tag or Refresh header method because
    they aren't predictable within Ajax (for example, JQuery seems to ignore them).
    '''
    def __init__(self, redirect_to=None, *args, **kwargs):
        # set up the code
        if redirect_to:
            script = 'window.location.assign("{}");'.format(redirect_to.split('#')[0])
        else:
            script = 'window.location.assign(window.location.href.split("#")[0])'
        # do we need to add the <script> tag? (that's the default)
        if kwargs.pop('include_script_tag', True):
            script = '<script>{}</script>'.format(script)
        # call the super
        super().__init__(script, *args, **kwargs)
        # add the custom header
        self[REDIRECT_HEADER_KEY] = redirect_to or 'window.location.href'
